#! /usr/bin/env python
# Copyright (C) 2010 by Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Build images and installers for a distribution."""

import glob, os, shutil, subprocess, sys, time

import bazaar_releases, buildoutgen, helpers, issgen


# Flag to enable dry-run mode. In this mode, file operations are done but
# commands aren't actually executed.
_DRY_RUN = False


DEFAULT_MSVC_REDIST_DIR = r'C:\Program Files\Microsoft Visual Studio 9.0' + \
    r'\VC\redist\x86\Microsoft.VC90.CRT'


## File helper functions ##

def chdir(dirname):
    print "Changing directory to %s" % (dirname,)
    if not _DRY_RUN:
        os.chdir(dirname)


def makedir(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    if not os.path.isdir(dirname):
        print "Error: Destination is not a directory", dirname
        return 2
    return 0


def copy_files(src, dest, files=None, verbose=False, overwrite=True):
    """Copy across files if they don't already exist.
    
    :param src: source directory
    :param dest: destination directory
    :param files: if None, all files in src are copied into dest
    """
    if files is None:
        files = os.listdir(src)
    for f in files:
        target = os.path.join(dest, f)
        target_dir = os.path.dirname(target)
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        if overwrite or not os.path.exists(target):
            original = os.path.join(src, f)
            if verbose:
                print "%s -> %s" % (original, target)
            shutil.copyfile(original, target)


def remove(files):
    """Remove files or directories."""
    filesdirs = glob.glob(files)
    for i in filesdirs:
        if os.path.isdir(i):
            shutil.rmtree(i)
            print "Removed:", i
        elif os.path.isfile(i):
            os.remove(i)
            print "Removed:", i
        else:
            print "Not found:", i


## Execution helper functions ##

_LOG_FILE = None

def _set_log_file(log_dir, log_id):
    timestamp = time.strftime("%Y%m%d-%H%M")
    log_basename = "%s-%s.log" % (log_id, timestamp)
    global _LOG_FILE
    _LOG_FILE = os.path.join(log_dir, log_basename)


def run_cmd(args, env=None, good_exit_codes=[0]):
    "Run a command, exiting if the result is an unexpected error code."""
    cmd = " ".join(args)
    print "Running %s ..." % (cmd,)
    if _DRY_RUN:
        return
    if env:
        # Ensure values are plain strings, not unicode
        for k, v in env.items():
            env[k] = str(v)
    if _LOG_FILE:
        # TODO: make this non-blocking
        proc = subprocess.Popen(args, env=env,
            stdout=subprocess.PIPE)
        log_file = open(_LOG_FILE, "a")
        try:
            output = proc.communicate()[0]
            log_file.write(output)
        finally:
            log_file.close()
        result = proc.returncode
    else:
        result = subprocess.call(args, env=env)
    if result not in good_exit_codes:
        sys.exit(result)


def run_bzr(args):
    if sys.platform == 'win32':
        # Use %COMSPEC% /c in case bzr is actually a .bat file
        comspec = os.environ['COMSPEC']
        cmd_args = [comspec, "/c", "bzr"]
    else:
        cmd_args = ["bzr"]
    run_cmd(cmd_args + args)
 

def python_setup(args, python="python", env=None):
    run_cmd([python, "setup.py"] + args, env=env)


## Builder classes ##

class BuildoutBuilder(object):

    def __init__(self, build_dir="build-win32", python="python"):
        self.build_dir = build_dir
        self.python = python

    def build(self, distros, names=None):
        """Build images for one or more distributions.

        :param distros: the list of distro objects
        :return: the build directory path
        """
        # Build the images
        self.init_build_dir()
        #_set_log_file(self.build_dir, "buildout")
        for distro in distros:
            self.build_image(distro)
        return os.getcwdu()

    def init_build_dir(self):
        """Create and initialise the build directory.
     
        On completion, the current directory is the build directory.
        """
        # Init build-win32
        print "Initialising %s directory" % (self.build_dir,)
        makedir(self.build_dir)
        copy_files("filesets/bootstrap", self.build_dir)

        # Bootstrap buildout
        os.chdir(self.build_dir)
        run_cmd([self.python, "bootstrap.py"])

    def build_image(self, distro):
        """Build an image.

        :param distro: a dict of projects, doc_sets and downloads.
           See bazaar_releases for examples.
        """
        # Use buildout to assemble the pieces into an image
        name = distro['name']
        print "Generating buildout configuration for %s ..." % (name,)
        bo_cfg = buildoutgen.generate_buildout_cfg(distro)
        bo_cfg = os.path.basename(bo_cfg)
        bo_cmd = os.path.join("bin", "buildout")
        run_cmd([bo_cmd, "-c", bo_cfg])


class WinHelpBuilder(object):

    def __init__(self, root, python="python"):
        """Create a Windows Help builder.

        :param root: the buildout root directory. See above for
            details on the expected layout.
        """
        self.root = root
        # For cygwin make, path names must have forward slashes
        self.python = python.replace('\\', '/')

    def build(self, distros):
        """Build Windows help for one or more distributions.

        :param distros: the list of distro objects
        """
        for distro in distros:
            self._build_help(distro)

    def _build_help(self, distro):
        # Build the help files and compile them
        distro_name = distro['name']
        print "\n### Building WinHelp for %s ###" % (distro_name,)
        chm_dir = "%s/chm-files/%s" % (self.root, distro_name)
        makedir(chm_dir)
        bzr_root = "%s/bzr/%s" % (self.root, distro_name)
        chdir(bzr_root)
        run_cmd(["make", "chm-sphinx", "PYTHON=%s" % self.python])
        for docset in self._docsets(distro):
            hhp_file = self._path_for(docset, "hhp")
            run_cmd(["hhc", hhp_file], good_exit_codes=[0, 1])
            chm_file = self._path_for(docset, "chm")
            chm_dest = os.path.join(chm_dir, os.path.basename(chm_file))
            print "copying %s -> %s" % (chm_file, chm_dest)
            if not _DRY_RUN:
                shutil.copyfile(chm_file, chm_dest)

    def _docsets(self, distro):
        """Get the docsets for a distro."""
        return [ds.id for ds in distro['docsets']]

    def _path_for(self, docset, ext):
        return "doc/%s/_build/htmlhelp/bzr-%s.%s" % (docset, docset, ext)


class InstallationSetBuilder(object):

    def __init__(self, root, compiler=None, python="python",
            msvc_redist_dir=DEFAULT_MSVC_REDIST_DIR):
        """Create an installation set builder.

        :param root: the buildout root directory. See above for
            details on the expected layout.
        :param compiler: the C compiler for the builds or None for the
            default one.
        :param python: the path to the Python executable.
        """
        self.root = os.path.abspath(root)
        self.compiler = compiler
        self.python = python
        self.msvc_redist_dir = msvc_redist_dir

    def build(self, distros, pythons=[]):
        """Build a set of installers for a set of distributions.

        A standalone installer is built along with zero or more
        Python-based ones. The latter include just the core software.
        The standalone installer adds help files, plugins and
        TortoiseBZR.

        Before running a build, a buildout directory needs to be
        populated. The expected directories are:

        * one for the distro-name, e.g. bzr-2.1
        * chm_files
        * excluding bzr, one for each project, e.g. qbzr
        * one for each download, e.g. tortoise_overlays, zlib

        Each project and download directory has <distro-name> subdirectrories.

        Installers are built under "<root>/releases/<distro-name>" and
        copied to the <root>/installers directory.

        :param distros: the list of distro objects
        :params pythons: the list of directories holding Python
            installations that Python-based installers should
            be built for.
        """
        for d in distros:
            self.build_distro(d, pythons)

    def build_distro(self, distro, pythons):
        self.distro = distro
        self.distro_name = distro['name']
        self.target = "%s/releases/%s" % (self.root, self.distro_name)
        self.zlib_base = "%s/zlib" % self.root

        # Log the build process (ENABLE LATER)
        log_dir = os.path.join(self.root, "releases")
        #_set_log_file(log_dir, self.distro_name)

        # Checkout the base branch
        remove(self.target)
        makedir(self.target)
        src_branch = "%s/bzr/%s" % (self.root, self.distro_name)
        run_bzr(["checkout", src_branch, self.target])

        # Build the Python installers first. We don't want to
        # include any of the 3rd-party plugins in these as
        # we don't want to bundle their dependencies.
        chdir(self.target)
        for py in pythons:
            self._bdist_wininst(py)

        # Build the various projects
        projects = [p.lp_project for p in distro['projects']]
        projects += [p.lp_project for p in distro['plugins'] if not p.core]
        dests = dict([(d.part_name, d.dest) for d in distro['downloads']])
        plugin_env = os.environ.copy()
        plugin_env.update({
            'SVN_DEV': os.path.join(self.root, dests["svn-dev"]),
            'SVN_BDB': os.path.join(self.root, dests["svn-db4"]),
            'SVN_LIBINTL': os.path.join(self.root, dests["svn-libintl"]),
            })
        for project in projects:
            if  project == 'bzr':
                # we do this last
                continue
            elif project == 'tortoisebzr':
                self._build_tortoisebzr()
            else:
                self._build_plugin(project, plugin_env)
        self._build_bzr()

        # Build the exe file and package the lot into an installer
        self._build_exe()
        self._build_standalone(self.distro_name)
        self._publish()

    def _bdist_wininst(self, python):
        python_setup([
            'bdist_wininst',
            '--install-script=bzr-win32-bdist-postinstall.py',
            '-d',
            '.',
            'build_ext'] + self._build_ext_args(), python)

    def _build_args(self):
        if self.compiler:
            result = ["-c", self.compiler]
        else:
            result = []
        return result

    def _build_ext_args(self):
        return self._build_args() + [
            "-I%s" % self.zlib_base,
            "-L%s" % self.zlib_base,
            ]

    def _build_tortoisebzr(self):
        print '\n### Building TortoiseBZR ###'
        tbzr = "%s/tortoisebzr/%s" % (self.root, self.distro_name)
        chdir(tbzr)
        python_setup(["build"] + self._build_args(), python=self.python,
            env=self._tbzr_env(tbzr))

    def _tbzr_env(self, tbzr):
        """Build the environment variables needed to build TortoiseBZR."""
        pypath = os.environ.get('PYTHONPATH')
        path = os.environ.get('PATH') or os.environ.get('Path')
        if pypath:
            pypath += os.pathsep + self.target
        else:
            pypath = self.target
        tbzr_base = os.path.basename(bazaar_releases._TBZR)
        tbzr_prefix = os.path.join(self.root, "tortoise-overlays", tbzr_base)
	env = os.environ.copy()
	env.update(dict(
            TBZR = tbzr,
            TORTOISE_OVERLAYS_MSI_WIN32 = tbzr_prefix + "win32.msi",
            TORTOISE_OVERLAYS_MSI_X64 = tbzr_prefix + "x64.msi",
            PYTHONPATH = pypath,
            PATH = path,
        ))
	return env

    def _build_plugin(self, name, env):
        print "\n### Installing %s ###" % (name,)
        in_dir = "%s/%s/%s" % (self.root, name, self.distro_name)
        chdir(in_dir)
        cmd_args = ["install", "-O1", "--install-lib=%s" % self.target]
        if name in ['subvertpy']:
            cmd_args.extend(self._build_args() +
                ["--install-scripts=%s" % self.target])
        python_setup(cmd_args, python=self.python, env=env)

    def _build_bzr(self):
        print '\n### Building the C extensions ###'
        chdir(self.target)
        remove("bzrlib/*.pyd")
        python_setup(["build_ext", "-i", "-f"] + self._build_ext_args(),
            python=self.python)

    def _build_exe(self):
        print 'Generating "bzr.exe"'
        tbzr = "%s/tortoisebzr/%s" % (self.root, self.distro_name)
        # TODO: capture the output to py2exe.log
        python_setup(["py2exe"], python=self.python,
            env=self._tbzr_env(tbzr))

    def _build_standalone(self, name):
        """Make the standlone installer itself."""
        # Copy across the required and help files
        exe_dir = "win32_bzr.exe"
        chm_dir = os.path.join(self.root, "chm-files", self.distro_name)
        copy_files(chm_dir,  exe_dir)
        inno_dir = os.path.join(self.root, "..", "filesets", "inno")
        copy_files(inno_dir, exe_dir)

        # Copy across the redistributable DLLs for Python 2.6.
        # This isn't required for Python 2.5 or earlier because
        # py2exe handles it then.
        if self.python.find("26") >= 0:
            msvc_base = os.path.basename(self.msvc_redist_dir)
            copy_files(self.msvc_redist_dir, exe_dir, [
                "msvcr90.dll",
                "%s.manifest" % msvc_base,
                ])

        # Generate the Installer spec file and compile it
        iss_file = issgen.generate_iss(self.distro,
            dest_dir=os.path.join("tools", "win32"))
        run_cmd(["iscc", "/Q", iss_file])

    def _publish(self):
        """Copy built installers into <root>/installers."""
        pattern = "%s/bzr-*.exe" % self.target.replace('\\', '/')
        files = glob.glob(pattern)
        relpaths = [os.path.basename(f) for f in files
            if not os.path.isdir(f)]
        if relpaths:
            dest = os.path.join(self.root, "installers")
            makedir(dest)
            print "\n### Copying installers into %s ###" % (dest,)
            copy_files(self.target, dest, relpaths, verbose=True)



def main():
    # Parse the command line
    from optparse import OptionParser
    usage = "usage: build [options] {parts|winhelp|installers} [distro-names]"
    parser = OptionParser(usage=usage)
    parser.add_option("--dry-run", action="store_true",
        help="Show run commands without executing them.",
        )
    parser.add_option("--standalone-only", action="store_true",
        help="Skip building the python-based compilers.",
        )
    parser.add_option("--compiler",
        help="Alternative compiler to the default, e.g. mingw32.",
        )
    parser.add_option("--pythons-dir",
        help="Base directory to look for Python installs in, e.g. D:\\.",
        )
    parser.add_option("--python",
        help="Directory within python-dirs for the default one, e.g. Python26.",
        )
    parser.add_option("--msvc-redist-dir",
        help="Directory holding MS Visual Studio redistribution files.",
        )
    parser.set_defaults(
        pythons_dir="c:\\",
        python="Python26",
        msvc_redist_dir=DEFAULT_MSVC_REDIST_DIR,
        )
    (options, args) = parser.parse_args()

    # Check usage
    if len(args) == 0:
        parser.error("Build type not specified.")
    build_type = args.pop(0)
    if build_type not in ['parts', 'winhelp', 'installers']:
        parser.error("Expected a build type, got %s." % (build_type,))
    distros = helpers.find_distros(args or None)

    # Do the requested action
    if options.dry_run:
        global _DRY_RUN
        _DRY_RUN = True
    build(build_type, distros, options)


def build(build_type, distros, options):
    """Build images or installers for one of more distributions.

    :param build_type: either parts, winhelp or installers
    """
    # Validate the parameters
    if build_type == 'parts':
        steps = ['buildout']
    elif build_type == 'winhelp':
        steps = ['buildout', 'winhelp']
    elif build_type == 'installers':
        steps = ['buildout', 'winhelp', 'installers']
    else:
        raise AssertionError("unknown build-type %s" % build_type)

    # Do the required steps
    build_dir = os.path.join(os.getcwdu(), "build-win32")
    cfg = get_configuration(options)
    for step in steps:
        if step == 'buildout':
            builder = BuildoutBuilder(build_dir, python=cfg['python'])
            builder.build(distros)
        elif step == 'winhelp':
            builder = WinHelpBuilder(build_dir, python=cfg['python'])
            builder.build(distros)
        elif step == 'installers':
            build_installers(distros, cfg, build_dir, options.standalone_only,
                options.msvc_redist_dir)


def get_configuration(options):
    """Get the list of Pythons, the default one and the compiler."""
    result = {}
    if options.compiler:
        result['compiler'] = options.compiler
    pythons = []
    if sys.platform == 'win32':
        pat = os.path.join(options.pythons_dir, "Python*")
        dirs = glob.glob(pat)
        for d in dirs:
            py_exe = os.path.join(d, "python.exe")
            basename = os.path.basename(d)
            if basename == options.python:
                result['python'] = py_exe
            pythons.append(py_exe)
    if not pythons:
        pythons = ['python']
    result['pythons'] = pythons
    if 'python' not in result:
        result['python'] = pythons[-1]
    return result
 

def build_installers(distros, cfg, build_dir, standalone_only,
        msvc_redist_dir):
    compiler = cfg.get('compiler')
    if standalone_only:
        pythons = []
    else:
        pythons = cfg['pythons']
    builder = InstallationSetBuilder(build_dir, compiler=compiler,
        python=cfg['python'], msvc_redist_dir=msvc_redist_dir)
    builder.build(distros, pythons)


if __name__ == '__main__':
    main()
