# Copyright (C) 2010 by Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Collect of common helper functions."""


import os


def find_distros(names):
    """Get the list of distributions matching a list of names.

    An illegal name outputs an error message and exits.
    If names is None, all distributions are returned.
    """
    import sys
    import bazaar_releases
    distros = dict([(d['name'], d) for d in bazaar_releases.DISTRIBUTIONS])
    all_names = sorted(distros)
    if names is not None:
        result = []
        for name in names:
            if name not in all_names:
                print "ERROR: distro %s not in known set: %s" % (name,
                    ", ".join(all_names))
                sys.exit(1)
            result.append(distros[name])
    else:
        result = bazaar_releases.DISTRIBUTIONS
    return result



def context_for_distro(distro, root_dir):
    """Get a context dictionary given a distro object and a root directory.
    
    :param distro: dictionary defining the pieces making up a distro.
    :param root_dir: directory containing the project's setup.py file, if any.
    :return: a dctionary containing at least these keys:
      * meta - a dictionary of meta information from setup.py
      * version - a version string
      * version_info - a tuple of version parts
    """
    result = distro.copy()
    result['meta'] = _get_setup_metadata(distro, root_dir)

    # Get the version. We use the one in bzrlib, if any.
    # If not, we try the meta information or fall back
    # to parsing the series from the distro name.
    version = None
    bzrlib_file = os.path.join(root_dir, "bzrlib", "__init__.py")
    if os.path.exists(bzrlib_file):
        import imp
        bzrlib = imp.load_source("bzrlib", bzrlib_file)
        version = bzrlib.__version__
        version_info = bzrlib.version_info
    else:
        try:
            version = result['meta']['version']
        except KeyError:
            version = distro['name'].split('-')[-1]
        version_info = version.split('.')
    result['version'] = version
    result['version_info'] = version_info
    return result


def _get_setup_metadata(distro, root_dir):
    result = {}
    # Get settings from setup.py if it exists
    setup_file = os.path.join(root_dir, "setup.py")
    if os.path.exists(setup_file):
        # XXX: if there a better way of doing this?
        # XXX: Hmm - it crashes with "invalid command name 'bzr-2.1'"
        # and I have *no* idea why :-( Maybe there's something weird
        # about bzr's setup.py file?
        import imp
        try:
            setup = imp.load_source("setup", setup_file)
            result = setup.META_INFO
        except:
            pass
    return result


def jinja_compile(template_file, context, output_file=None):
    """Compile a template using a context and optionally output the result.

    :param template_file: the file containing the template
    :param context: dictionary of context objects
    :param output_file: if not None, the file to write to
    :return: the result
    """
    from jinja2 import Environment, FileSystemLoader
    def save(content, filename):
        f = open(filename, 'w')
        try:
            f.write(content.encode('utf-8'))
        finally:
            f.close()
    template_dir, file = os.path.split(template_file)
    loader = FileSystemLoader(template_dir)
    env = Environment(loader=loader)
    template = env.get_template(file)
    output = template.render(context)
    if output_file:
        save(output, output_file)
    return output


def jinja_compile_dir(context, template_dir, output_dir, files=None):
    """Compile templates using a context and output the results.

    :param context: dictionary of context objects
    :param template_dir: the directory to search for templates in
    :param output_dir: the directory to save results in.
        (It is currently assumed that this directory exists.)
    :param files: the set of templates to compile. If None, all
        templates are compiled.
    :result: the paths to the files output
    """
    from jinja2 import Environment, FileSystemLoader
    def save(content, filename):
        f = open(filename, 'w')
        try:
            f.write(content.encode('utf-8'))
        finally:
            f.close()
    loader = FileSystemLoader(template_dir)
    env = Environment(loader=loader)
    if files is None:
        files = os.listdir(template_dir)
    result = []
    for file in files:
        template = env.get_template(file)
        output = template.render(context)
        path = os.path.join(output_dir, file)
        save(output, path)
        result.append(path)
    return result
