#! /usr/bin/env python
# Copyright (C) 2010 by Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Build an installer specification file for a distribution."""


import os

import bazaar_releases, helpers


def generate_iss(distro, dest_dir='.', root_dir="."):
    """Generate an iss file for a distribution.

    :param distro: the distro object. See bazaar_releases for examples.
    :param dest_dir: the destination directory
    :return: the name of the generated file.
    """
    iss_file = "bzr.iss"
    print "Generating %s in %s" % (iss_file, dest_dir)
    this_dir = os.path.dirname(__file__)
    template_dir = os.path.join(this_dir, "templates", "inno")
    context = helpers.context_for_distro(distro, root_dir)
    #print "context: %s" % (context,)
    context.update(_get_special_settings(context))
    build_dir = os.path.abspath(os.path.join(this_dir, "build-win32"))
    context.update(_get_tbzr_settings(distro, build_dir))
    return helpers.jinja_compile_dir(context, template_dir, dest_dir,
        [iss_file])[0]


def _get_special_settings(context):
    """Return a dictionary of special settings.
    
    These include:
    
    * VersionInfoVersion - a string in the form a.b.c.d where each part is a number.
    """
    version_info = context['version_info']
    if len(version_info) >= 3:
        numbers = [int(p) for p in version_info[:3]]
    else:
        numbers = [int(p) for p in version_info] + [0, 0, 0]
    for i in range(0, 3):
        numbers[i] = str(numbers[i])
    numbers = numbers[:3] + [str(_get_build_number())]
    viv = ".".join(numbers)
    return {'VersionInfoVersion': viv}


def _get_build_number():
    # TODO: calculate the build number by looking at builds already made
    return 1


def _get_tbzr_settings(distro, build_dir):
    """Get special TortoiseBZR settings needed by the installer."""
    olay_basename = os.path.basename(bazaar_releases._TBZR)
    olay_prefix = os.path.join(build_dir, "tortoise-overlays", olay_basename)
    result = {}
    result['tbzr'] = os.path.join(build_dir, "tortoisebzr", distro['name'])
    result['tbzr_msi32'] = '%s%s.msi' % (olay_prefix, 'win32')
    result['tbzr_msi64'] = '%s%s.msi' % (olay_prefix, 'x64')
    result['tbzr_msi32_basename'] = '%s%s.msi' % (olay_basename, 'win32')
    result['tbzr_msi64_basename'] = '%s%s.msi' % (olay_basename, 'x64')
    return result


if __name__ == '__main__':
    import sys
    if len(sys.argv[1:]) == 0:
        print "usage: issgen distro-name [root-dir]"
        sys.exit(1)
    distro = helpers.find_distros([sys.argv[1]])[0]
    try:
        root_dir = sys.argv[2]
    except IndexError:
        root_dir = '.'
    generate_iss(distro, root_dir=root_dir)
